package ru.dbotthepony.mc.otm.core.util

import net.minecraft.core.Direction
import net.minecraft.core.SectionPos
import net.minecraft.world.level.block.entity.BlockEntity
import net.minecraftforge.common.capabilities.Capability
import net.minecraftforge.common.util.LazyOptional
import ru.dbotthepony.mc.otm.core.collect.WeakHashSet
import ru.dbotthepony.mc.otm.core.math.plus
import ru.dbotthepony.mc.otm.core.orNull
import ru.dbotthepony.mc.otm.onceServer
import java.lang.ref.WeakReference
import java.util.EnumMap
import java.util.function.Predicate

open class BESubscribeList<T>(
	val block: BlockEntity,
	val capability: Capability<T>,
) : Predicate<Direction>, Iterable<T> {
	private val knownCaps = WeakHashSet<LazyOptional<T>>()
	private val trackedCaps = EnumMap<Direction, WeakReference<LazyOptional<T>>>(Direction::class.java)

	var valid = true
		set(value) {
			if (value != field) {
				field = value

				if (value) {
					update()
				}
			}
		}

	fun invalidate() {
		valid = false
	}

	fun revive() {
		valid = true
	}

	// override
	protected open fun subscribe(direction: Direction, capability: LazyOptional<T>) {}
	// override
	protected open fun unsubscribe(direction: Direction, capability: LazyOptional<T>?) {}

	override fun test(t: Direction): Boolean {
		return true
	}

	fun validate() {
		val iterator = trackedCaps.iterator()

		for ((k, v) in iterator) {
			if (v.get() == null || !test(k)) {
				unsubscribe(k, null)
			}
		}
	}

	override fun iterator(): Iterator<T> {
		return object : Iterator<T> {
			val iterator = trackedCaps.iterator()
			var value: T? = null

			private fun find() {
				while (iterator.hasNext() && value == null) {
					val (k, v) = iterator.next()
					value = v.get()?.orNull()

					if (value == null || !test(k)) {
						iterator.remove()
						unsubscribe(k, null)
					}
				}
			}

			override fun hasNext(): Boolean {
				find()
				return value != null
			}

			override fun next(): T {
				find()
				val value = value
				this.value = null
				return value ?: throw NoSuchElementException()
			}
		}
	}

	fun update(direction: Direction) {
		return update(direction::equals)
	}

	private var updateTick = 0

	/**
	 * Updates subscription list, searching for new capabilities or "removing" outdated ones
	 *
	 * [predicate] predicate narrows sides-to-check list, not replaces list generated by this class' [test]
	 */
	@JvmOverloads
	fun update(predicate: Predicate<Direction>? = null) {
		if (!valid)
			return

		val sorse = block.level?.chunkSource ?: return
		var waitChunkLoad = false

		updateTick++
		val updateTick = updateTick
		var stream = Direction.stream().filter(this)

		if (predicate != null) {
			stream = stream.filter(predicate)
		}

		for (dir in stream) {
			val pos = block.blockPos + dir
			val getChunk = sorse.getChunkNow(SectionPos.blockToSectionCoord(pos.x), SectionPos.blockToSectionCoord(pos.z))

			if (getChunk == null) {
				waitChunkLoad = true
			} else {
				val cap = getChunk.getBlockEntity(pos)?.getCapability(capability)
				val knownCap = trackedCaps[dir]?.get()

				if (cap != null && cap.isPresent) {
					if (knownCap !== cap) {
						if (knownCaps.add(cap)) {
							val ref = WeakReference(this)

							cap.addListener {
								val self = ref.get()

								if (self != null) {
									val current = self.trackedCaps[dir]

									if (current?.get() === it) {
										self.update()
									}
								}
							}
						}

						trackedCaps[dir] = WeakReference(cap)
						subscribe(dir, cap)
					}
				} else {
					trackedCaps.remove(dir)
					unsubscribe(dir, knownCap)
				}
			}
		}

		if (waitChunkLoad) {
			onceServer { if (updateTick == this.updateTick) update(predicate) }
		}
	}
}
